from itertools import takewhile
from statistics import mean, StatisticsError

from ndnt.line import Line
from ndnt.lines import Lines


class Indent:
    """Line's indent."""

    def __init__(self, origin: Line):
        self.origin = origin

    def value(self) -> str:
        """Indent itself."""
        return "".join(takewhile(str.isspace, self.origin))

    def length(self) -> int:
        """Length of indent in spaces.

        Each tab counts as 4 spaces.
        """
        return len(self.value().replace("\t", " " * 4))


class AverageIndent:
    """Average indent of lines block."""

    def __init__(self, origin: Lines):
        self.origin = origin

    def value(self) -> float:
        """Average indent itself."""
        indents = map(Indent, self.origin)
        try:
            return mean(indent.length() for indent in indents)
        except StatisticsError:
            # if origin has no lines
            return 0.0


class RoundedAverageIndent:
    """Rounded `AverageIndent`."""

    def __init__(self, origin: AverageIndent):
        self.origin = origin

    def value(self) -> float:
        """Rounded average indent itself."""
        return round(self.origin.value(), ndigits=2)



from pathlib import Path
from typing import Iterable

from pathspec import PathSpec
from pathspec.patterns import GitWildMatchPattern


class Paths(Iterable[Path]):
    """Paths inteface.

    `Paths` behaves like `Iterable[Path]`.
    """


class FilesPaths(Paths):
    """Files paths.

    All paths that are files from given direcatory.
    """

    def __init__(self, path: Path):
        self.path = path

    def __iter__(self) -> Iterable[Path]:
        return filter(Path.is_file, self.path.glob("**/*"))


class ExtensionPaths(Paths):
    """Extension paths.

    Paths with certain extension.
    """

    def __init__(self, origin: Paths, extension: str):
        self.origin = origin
        self.extension = extension

    def __iter__(self) -> Iterable[Path]:
        return filter(lambda path: path.suffix == self.extension, self.origin)


class ExcludeGitignoredPaths(Paths):
    """Paths exluding paths that matches gitignore.

    Optimization inspired by "black" package:
    https://github.com/psf/black/blob/fda2561f79e10826dbdeb900b6124d642766229f/src/black/files.py#L177
    """

    def __init__(self, path: Path, gitignore_path: Path):
        self.path = path
        self.gitignore_path = gitignore_path

    def __iter__(self) -> Iterable[Path]:
        gitignore = PathSpec.from_lines(
            GitWildMatchPattern, self.gitignore_path.read_text().splitlines()
        )
        yield from self._iter(self.path, gitignore)

    def _iter(self, path: Path, gitignore: PathSpec) -> Iterable[Path]:
        """Optimized gitignore paths filter.

        Do not iterate directories if they are in gitignore.
        """
        for p in path.iterdir():
            if gitignore.match_file(p):
                continue

            if p.is_file():
                yield p
            elif p.is_dir():
                yield from self._iter(p, gitignore)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable

from ndnt.indent import AverageIndent, RoundedAverageIndent
from ndnt.lines import LinesFromFile, LinesFromFiles, NonBlankLines


class Summary(ABC):
    """Summary interface.

    All `Summary` subclasses must have `print()` method to display its summary.
    """

    @abstractmethod
    def print(self): ...


class FileSummary(Summary):
    """File summary.

    Shows file's average indent (excluding blank lines) and its path.
    """

    def __init__(self, path: Path):
        self.path = path

    def print(self):
        lines = NonBlankLines(LinesFromFile(self.path))
        indent = RoundedAverageIndent(AverageIndent(lines)).value()
        print(f"{indent:<5} |", self.path)


class FilesSummary(Summary):
    """Directory summary.

    Shows average indent of multiple files.
    """

    def __init__(self, paths: Iterable[Path]):
        self.paths = paths

    def print(self):
        total_lines = NonBlankLines(LinesFromFiles(self.paths))
        total_indent = RoundedAverageIndent(AverageIndent(total_lines)).value()
        print(f"{total_indent:<5} | Total")


class DirectorySummary(Summary):
    """Directory summary.

    Shows average indent of all files separately and then
    average indent of all files combined.
    """

    def __init__(self, paths: Iterable[Path]):
        self.paths = paths

    def print(self):
        for file in self.paths:
            FileSummary(file).print()

        print("-" * 13)
        FilesSummary(self.paths).print()



from pathlib import Path
from typing import List
from setuptools import find_packages, setup


CURRENT_DIR = Path(__file__).parent


def get_long_description() -> str:
    """Get long description from README.md."""
    return (CURRENT_DIR / "README.md").read_text(encoding="u8")


def get_install_requires() -> List[str]:
    """Get duplicated requirements from requirements.txt."""
    return ["pathspec==0.9.0"]


setup(
    name="ndnt",
    version="1.1.0",
    description="Inspect indents of your files.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/Masynchin/ndnt",
    author_email="masynchin@gmail.com",
    license="MIT",
    packages=find_packages(include=["ndnt"]),
    python_requires=">=3.7",  # I am not sure about it
    install_requires=get_install_requires(),
    entry_points={
        "console_scripts": [
            "ndnt=ndnt.__main__:main",
        ]
    },
)



from itertools import chain
from pathlib import Path
from typing import Iterable

from ndnt.line import Line


class Lines(Iterable[Line]):
    """Lines interface.

    `Lines` behaves like `Iterable[Line]`.
    """


class LinesFromText(Lines):
    """Lines from text."""

    def __init__(self, text: str):
        self.text = text

    def __iter__(self) -> Iterable[Line]:
        return map(Line, self.text.splitlines())


class LinesFromFile(Lines):
    """Lines from file."""

    def __init__(self, path: Path):
        self.path = path

    def __iter__(self) -> Iterable[Line]:
        yield from LinesFromText(self.path.read_text(encoding="u8"))


class NonBlankLines(Lines):
    """Non blank lines."""

    def __init__(self, origin: Lines):
        self.origin = origin

    def __iter__(self) -> Iterable[Line]:
        return filter(None, self.origin)


class CombinedLines(Lines):
    """Multiple `Lines` combined as one `Lines`."""

    def __init__(self, origins: Iterable[Lines]):
        self.origins = origins

    def __iter__(self) -> Iterable[Line]:
        return chain.from_iterable(self.origins)


class LinesFromFiles(Lines):
    """Lines from multiple files as one `Lines`."""

    def __init__(self, paths: Iterable[Path]):
        self.paths = paths

    def __iter__(self) -> Iterable[Line]:
        yield from CombinedLines(map(LinesFromFile, self.paths))



from argparse import ArgumentParser
from pathlib import Path

from ndnt.paths import ExcludeGitignoredPaths, ExtensionPaths, FilesPaths
from ndnt.summary import DirectorySummary, FileSummary


def setup_parser() -> ArgumentParser:
    """Setup CLI parser."""
    parser = ArgumentParser(description="Inspect indents of your files.")
    parser.add_argument(
        "path", type=Path, help="Whether file or directory to get summary of."
    )
    parser.add_argument(
        "-e",
        "--extension",
        default=".py",
        help="Filter files by extension.",
    )
    parser.add_argument(
        "--no-gitignore",
        action="store_true",
        help="Do not exclude paths matches gitignore.",
    )
    return parser


def parse_args():
    """Parse args provided by CLI."""
    return setup_parser().parse_args()


def main():
    """Main function called with `ndnt` command."""
    main_with_args(**parse_args().__dict__)


def main_with_args(path: Path, no_gitignore: bool, extension: str):
    """Main function.

    Choose summary depends on provided arguments and print it.
    """
    no_gitignore = no_gitignore or not (path / ".gitignore").exists()

    if path.is_file():
        summary = FileSummary(path)
    elif path.is_dir() and no_gitignore:
        summary = DirectorySummary(
            ExtensionPaths(FilesPaths(path), extension)
        )
    elif path.is_dir():
        summary = DirectorySummary(
            ExtensionPaths(
                ExcludeGitignoredPaths(path, path / ".gitignore"),
                extension,
            )
        )
    else:
        print(f"Something is wrong with provided path.")
        return

    summary.print()
